{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- Notes on Semigroup/Monoid, Functor/Applicative/Monad, Alternative/MonadPlus

module Juli8 (
    module Data.Semigroup
  , module Data.Functor
  , module Control.Applicative
  , module Control.Monad
  , module Data.Function
  , module Data.Tuple
  , module Data.Maybe
  , module Data.Either
  , module Control.DeepSeq
  , module Control.Exception
  , module System.IO.Unsafe
  , OptFirst, getOptFirst
  , OptLast, getOptLast
  , OptMin, getOptMin
  , OptMax, getOptMax
  ) where

import Data.Semigroup
import Data.Functor
import Control.Applicative
import Control.Monad
import Data.Function (fix, on)
import Data.Tuple (swap)
import Data.Maybe (isJust, isNothing, fromMaybe, fromJust, listToMaybe, maybeToList, mapMaybe, catMaybes)
import Data.Either (lefts, rights, isLeft, isRight, partitionEithers)
import Control.DeepSeq (deepseq, ($!!), force)
import Control.Exception (assert)
-- Prelude.undefined
-- Prelude.error "message"
-- Prelude.ioError $ Prelude.userError "message"; \e -> System.IO.Error.ioeGetErrorString e
import System.IO.Unsafe (unsafePerformIO)
-- should declare `# NOINLINE foo #` on any `foo` that invokes
-- should :set -fno-cse
-- if invocation could float outside a lambda, should :set -fno-full-laziness

{-

(*) `Data.Function.on` g = \x y -> g x * g y
Data.Tuple.swap (a,b) ==> (b,a)
Prelude.maybe defaultb (a->b) a? -> b
Data.Maybe.fromMaybe defaulta a? -> a
Data.Maybe.fromJust -- else error
Prelude.head/tail [] are errors
Data.Maybe.listToMaybe -- head?
Data.Maybe.maybeToList
Data.Maybe.mapMaybe (a->b?) [a] -> [b]
Data.Maybe.catMaybes [a?] -> [a] -- compare concatMap
Prelude.either (a->c) (b->c) ab? -> c
Data.Either.lefts/rights [ab?] -> [a]/[b]
Data.Either.partitionEithers [ab?] -> ([a],[b])




Data.SEMIGROUP (<> associative, sconcat (Data.List.NonEmpty = semi:|[semi]) -> a using <>)
-- re-exports overridden Data.Monoid declarations
times1p (Natural n) semi, replicate n+1 times using <>
timesN (Natural n) Monoid, replicate n times using <>
cycle1 semi, using <>
option defaultb (Semi->b) (Option semi) -> b
diff semi = Endo (semi <>), lifts any Semigroup into Monoid
Min/Max/First/Last(..), WrappedMonoid(..), Option(..), Arg(..), ArgMin/ArgMax

Data.MONOID (mappend/<>, mempty, mconcat [Mo])
Dual(..), Endo(..), All/Any/Sum/Product(..), Data.Monoid.First/Last(..)

-- All listed Monoids are also Semigroups
Monoid ()
Semigroup (Semigroup, Semigroup), pointwise; also Monoid (Monoid, Monoid)
Monoid Ordering where LT <> _ -> LT, EQ is mempty, GT <> _ -> GT (lexicographic)
Semigroup (First/Last { getFirst :: a }), always takes first/last operand
Semigroup (Const { getConst :: Semigroup } phantom) where Const a <> Const b = Const (a <> b); also Monoid (Const Monoid phantom)
Semigroup (Maybe Semigroup) where Just a <> Just b = Just (a <> b)
Monoid (All/Any { getAll :: Bool }) where All p <> All q = All (p && q)
Monoid (Sum/Product { getSum :: Num }) where Sum x <> Sum y = Sum (x + y)

Semigroup (Min/Max { getMin :: Ord }); if Bounded then Monoid with max/minBound
	Min/Max (Arg x g), Eq/Ord target the x, fmap targets the g
	case getMax $ foldr1 (<>) $ (\x -> Max (Arg x (10-))) <$> [1,4,2,3] of Arg x g -> g x ==> 6
	ArgMin/ArgMax a g = Min (Arg a g)
Monoid (OptMin/OptMac { getOptMin :: Maybe Ord }), see below

Monoid (Option { getOption :: Maybe Semigroup }), lifts any Semigroup into Monoid
	whereas Data.Monoid requires base to be Monoid, and ignores Just mempty
	Data.Monoid's First/Last a = Option (First a), first non-Nothing, else last (like Alternative's <|>)
Monoid (OptFirst/OptLast { getOptFirst :: Maybe Semigroup }), see below

Semigroup (Either a b), first non-Left, else last (like Alternative's <|>)

Monoid [a] where <> = ++
-- also Seq a, Text
Semigroup (NonEmpty a) where (a :| as) <> ~(b :| bs) = a :| (as ++ b : bs)
Monoid (Set Ord) using mappend
Monoid (Map Ord v) using mappend
-- also IntSet, IntMap v

Monoid (Endo (a->a)) where Endo g <> Endo h = Endo (g . h); mempty = id
Semigroup (a -> Semigroup), where g <> h = \a -> g a <> h a; also Monoid (a -> Monoid)
Semigroup (Dual Semigroup) where Dual a <> Dual b = Dual (b <> a); also Monoid (Dual Monoid)
Monoid (WrappedMonoid Monoid) using mappend




Data.FUNCTOR (fmap in Prelude, <$ = fmap.const in Data.Functor or Control.Applicative)
	fmap id = id
	fmap (g . h) = fmap g . fmap h
	fmap g Monad = Monad >>= return . g
<$> in Data.Functor or Control.Applicative
$> = flip (<$)
void = fmap (const ()) in Data.Functor or Control.Monad

Control.APPLICATIVE (pure/Prelude.return, <*>/Control.Monad.ap, *>/Prelude.>> = liftA2 (const id), <* = liftA2 const)
	pure g <*> pure x = pure (g x)
	pure id <*> v = v
	pure (.) <*> u <*> v <*> w = u <*> (v <*> w)
	pure ($ y) <*> u = u <*> pure y
entails: fmap g u = pure g <*> u
<**> = flip <*>
liftA, liftA2, liftA3, <$>, <$
Const(..), ZipList(..), WrappedMonad(..)

Control.MONAD (>>=, >>, return, fail, =<<, sequence, mapM all in Prelude)
        (liftM[2-5], ap, join = (>>= id), >=>, <=<, void in Control.Monad
sequence :: [M a] -> M [a]
mapM k [a] = sequence . map k [a]
when/unless Bool (M ()) = if Bool then M () else return ()
forever (u :: M a) = let u' = u >> u' in u'
forM[_], filterM, mapAndUnzipM, zipWithM[_], foldM[_], replicateM[_]

Monad Maybe
Monad Option
Monad (Either e)
Monad Min/Max/First/Last
Monad OptMin/OptMax/OptFirst/OptLast, see below
Functor (Arg a)
Monad [] where fmap = map, return = singleton, u >>= k = concatMap k u
Applicative (ZipList { getZipList :: [a] }) where pure a = repeat a, gs <*> xs = zipWith id gs xs
Monad NonEmpty where return a = a:|[], ~(a:|as) >>= k = let b:|bs = k a in b:|(bs ++ (as >>= toList.k))

Monad (r->) where fmap g (r->) = g . (r->), <*> = S combinator, return = const, u >>= k = \r -> k (u r) r
Applicative (ST s)
Functor (a,) where fmap g (a, b) = (a, g b); also Applicative (Monoid,) where pure x = (mempty, x), (u,g) <*> (v,x) = (u <> v, g x)
Functor (Const { getConst :: a }) where fmap _ (Const a) = Const a
Applicative (Const Monoid) where pure _ = Const mempty, Const u <*> Const v = Const (u <> v)
Applicative IO
Monad (WrappedMonad { unwrapMonad :: Monad a }), Alternative if Monad is MonadPlus
Monad Data.Functor.Identity { runIdentity :: a } -- [Control.Monad.Identity exports along with Control.Monad, Control.Monad.Fix]
Functor Data.Functor.Product.Pair (Functor a) (Functor' a), pointwise, MonadPlus if Functors are
Functor Data.Functor.Constant { getConstant :: a } phantom, Applicative if a :: Monoid
Functor Data.Functor.Compose { getCompose :: Functor (Functor' a) }, Applicative if Functors are, Alternative if outer is also Alternative
Functor Data.Functor.Reverse { getReverse :: Functor a }, Foldable and Traversable go in reverse order; Applicative/Alternative if Functor is




Data.Applicative.ALTERNATIVE (empty = MonadPlus's mzero, <|> = MonadPlus's mplus)
some (u :: Alternative a) = fmap (:) u <*> (some u <|> pure []) -- one or more as
many (u :: Alternative a) = (fmap (:) u <*> many u) <|> pure [] -- zero or more as
optional (u :: Alternative a) = fmap Just u <|> pure Nothing

MONADPLUS (mzero, associative mplus)
	mzero identity for mplus
	mzero >>= k == mzero, aborting further processing of this component (see mfilter)
	u >> mzero == mzero
guard Bool :: MP () or mzero
mfilter p (MP a) = MP a >>= \a -> if p a then return a else mzero
msum [MP a] -> MP a, using mplus and mzero

MonadPlus Maybe, first non-Nothing, else last
MonadPlus Option, first non-Nothing, else last
MonadPlus OptMin/OptMax/OptFirst/OptLast, see below
MonadPlus [] where empty/mzero = [], <|>/mplus = ++
MonadPlus IO/IOFirst/IOLast, see below

-}



-- OptFirst and OptLast are like Data.Monoid.First/Last
-- The Alternative/MonadPlus instances for OptFirst/OptLast/OptMin/OptMax stop at first non-Nothing
-- Only Semigroup and Monoid instances honor the First/Last/Min/Max behavior

newtype OptFirst a = OptFirst (Option (First a)) deriving (Eq, Ord, Read, Show, Semigroup, Monoid)
instance Functor OptFirst where { fmap g (OptFirst u) = OptFirst $ fmap g <$> u }
instance Applicative OptFirst where { pure a = OptFirst $ First <$> pure a; OptFirst u <*> OptFirst v = OptFirst $ liftA2 (<*>) u v }
instance Monad OptFirst where { return = pure; OptFirst u >>= k = let get (OptFirst a) = a in OptFirst (u >>= get . k . getFirst) }
instance Alternative OptFirst where { empty = OptFirst empty; OptFirst u <|> OptFirst v = OptFirst $ u <|> v }
instance MonadPlus OptFirst where { mzero = empty; mplus = (<|>) }
getOptFirst :: OptFirst a -> Maybe a
getOptFirst (OptFirst u) = getFirst <$> getOption u

newtype OptLast a = OptLast (Option (Last a)) deriving (Eq, Ord, Read, Show, Semigroup, Monoid)
instance Functor OptLast where { fmap g (OptLast u) = OptLast $ fmap g <$> u }
instance Applicative OptLast where { pure a = OptLast $ Last <$> pure a; OptLast u <*> OptLast v = OptLast $ liftA2 (<*>) u v }
instance Monad OptLast where { return = pure; OptLast u >>= k = let get (OptLast a) = a in OptLast (u >>= get . k . getLast) }
instance Alternative OptLast where { empty = OptLast empty; OptLast u <|> OptLast v = OptLast $ u <|> v }
instance MonadPlus OptLast where { mzero = empty; mplus = (<|>) }
getOptLast :: OptLast a -> Maybe a
getOptLast (OptLast u) = getLast <$> getOption u

newtype OptMin a = OptMin (Option (Min a)) deriving (Eq, Ord, Read, Show, Semigroup, Monoid)
instance Functor OptMin where { fmap g (OptMin u) = OptMin $ fmap g <$> u }
instance Applicative OptMin where { pure a = OptMin $ Min <$> pure a; OptMin u <*> OptMin v = OptMin $ liftA2 (<*>) u v }
instance Monad OptMin where { return = pure; OptMin u >>= k = let get (OptMin a) = a in OptMin (u >>= get . k . getMin) }
instance Alternative OptMin where { empty = OptMin empty; OptMin u <|> OptMin v = OptMin $ u <|> v }
instance MonadPlus OptMin where { mzero = empty; mplus = (<|>) }
getOptMin :: OptMin a -> Maybe a
getOptMin (OptMin u) = getMin <$> getOption u

newtype OptMax a = OptMax (Option (Max a)) deriving (Eq, Ord, Read, Show, Semigroup, Monoid)
instance Functor OptMax where { fmap g (OptMax u) = OptMax $ fmap g <$> u }
instance Applicative OptMax where { pure a = OptMax $ Max <$> pure a; OptMax u <*> OptMax v = OptMax $ liftA2 (<*>) u v }
instance Monad OptMax where { return = pure; OptMax u >>= k = let get (OptMax a) = a in OptMax (u >>= get . k . getMax) }
instance Alternative OptMax where { empty = OptMax empty; OptMax u <|> OptMax v = OptMax $ u <|> v }
instance MonadPlus OptMax where { mzero = empty; mplus = (<|>) }
getOptMax :: OptMax a -> Maybe a
getOptMax (OptMax u) = getMax <$> getOption u

